"""
Reinforcement learning environment interface
"""
import typing
from functools import partial

import chex
import jax

from esquilax.typing import TEnvParams, TEnvState


class Environment(typing.Generic[TEnvState, TEnvParams]):
    """
    RL environment interface for Esquilax simulations

    Basic abstract RL environment intended for use with
    built-in training and testing functionality
    (see :py:mod:`esquilax.ml.rl.training`).
    """

    def default_params(self) -> TEnvParams:
        """
        Get default environment parameters

        Returns
        -------
        esquilax.typing.TEnvParams
            Environment parameters.
        """
        raise NotImplementedError

    @partial(jax.jit, static_argnums=(0,))
    def reset(
        self, key: chex.PRNGKey, params: TEnvParams
    ) -> typing.Tuple[chex.Array, TEnvState]:
        """
        Reset the state of the environment

        Reset the environment and return the new state and
        agent observations.

        Parameters
        ----------
        key
            JAX random key.
        params
            Environment parameters.

        Returns
        -------
        tuple[jax.numpy.ndarray, esquilax.typing.TEnvState]
            Tuple containing

            - Agent observations, i.e. each agents
              observation of the sim state. If multiple
              agents, then the observation should have
              the same tree structure.
            - Environment state.
        """
        raise NotImplementedError

    @partial(jax.jit, static_argnums=(0,))
    def step(
        self,
        key: chex.PRNGKey,
        params: TEnvParams,
        state: TEnvState,
        actions: chex.ArrayTree,
    ) -> typing.Tuple[chex.ArrayTree, TEnvState, chex.ArrayTree, chex.ArrayTree]:
        """
        Update the state of the environment given agent actions

        Parameters
        ----------
        key
            JAX random key.
        params
            Environment parameters
        state
            Current environment state
        actions
            Agent actions, for multiple agents this could be
            a PyTree of arrays.

        Returns
        -------
        tuple
            Tuple containing:

            - Agent observations
            - New environment state
            - Agent rewards
            - Terminal flags

            In the case of multiple agents, the tree structure
            of each field should match the tree structure of the
            actions.
        """
        raise NotImplementedError



"""
Policy training patterns for use with `Evosax`_ neuro-evolution package

Classes definitions and functionality for training of evolutionary
policies (via `Evosax`_) using esquilax as a multi-agent
(and potentially multi-policy) training environment.

.. _Evosax: https://github.com/RobertTLange/evosax

.. note::

   This module requires additional dependencies that can be installed
   using the installation extra ``evo``, for example with pip using

   .. code::

      pip install esquilax[evo]
"""
try:
    import evosax  # noqa: F401
except ImportError as e:
    raise ImportError(
        "Additional requirements are required for this functionality. "
        "They can be installed with the extra `pip install esquilax[evo]`"
    ) from e
from . import tree_utils
from .strategy import BasicStrategy, Strategy
from .training import test, train
from .types import TrainingData



"""
RL agent interfaces and state
"""
from functools import partial
from typing import Tuple, Union

import chex
import jax

from .agent_state import AgentState, BatchAgentState


class Agent:
    """
    Abstract rl-agent class

    Outlines functionality required for agents to used
    as part of esquilax rl training functionality
    (see :py:mod:`esquilax.ml.rl.training`).
    """

    @partial(jax.jit, static_argnames=("self",))
    def sample_actions(
        self,
        key: chex.PRNGKey,
        agent_state: Union[AgentState, BatchAgentState],
        observations: chex.Array,
        greedy: bool = False,
    ) -> Tuple[chex.ArrayTree, chex.ArrayTree]:
        """
        Sample actions given observations

        Parameters
        ----------
        key
            JAX random key.
        agent_state
            Current agent training state.
        observations
            Environment observations.
        greedy
            Flag to indicate greedy action sampling.

        Returns
        -------
        tuple[chex.ArrayTree, chex.ArrayTree]
            Tuple containing actions and
            any corresponding action values
        """
        raise NotImplementedError

    @partial(jax.jit, static_argnames=("self",))
    def update(
        self,
        key: chex.PRNGKey,
        agent_state: Union[AgentState, BatchAgentState],
        trajectories,
    ) -> Tuple[Union[AgentState, BatchAgentState], chex.ArrayTree]:
        """
        Update the state of the agent from observed trajectories

        Parameters
        ----------
        key
            JAX random key.
        agent_state
            Current agent training state.
        trajectories
            Struct of environment update trajectories.

        Returns
        -------
        tuple[AgentState | BatchAgentState, chex.ArrayTree]
            Tuple containing an updated ``Agent`` and any associated data
            e.g. training losses from the update.
        """
        raise NotImplementedError



from functools import partial
from math import floor, isclose, prod
from typing import Any, Callable, Optional, Sequence, Tuple, Union

import chex
import jax
import jax.numpy as jnp

from esquilax import utils
from esquilax.reductions import Reduction
from esquilax.typing import Default


def _sort_agents(
    n_bins: Tuple[int, int], width: float, pos: chex.Array, agents: chex.Array
) -> Tuple[chex.Array, chex.Array, chex.Array, chex.Array, chex.Array, chex.ArrayTree]:
    co_ords, idxs = utils.space.get_bins(pos, n_bins, width)
    sort_idxs = jnp.argsort(idxs)
    _, bins = utils.graph.index_bins(idxs, prod(n_bins))
    sorted_co_ords = co_ords[sort_idxs]
    sorted_idxs = idxs[sort_idxs]
    sorted_pos = pos[sort_idxs]
    sorted_agents = jax.tree_util.tree_map(lambda y: y[sort_idxs], agents)

    return sorted_co_ords, sort_idxs, bins, sorted_idxs, sorted_pos, sorted_agents


def _process_parameters(
    i_range: float,
    dims: Union[float, Sequence[float]],
    n_bins: Optional[int | Sequence[int]],
) -> Tuple[Tuple[int, int], float, chex.Array]:
    if isinstance(dims, Sequence):
        assert (
            len(dims) == 2
        ), f"2 spatial dimensions should be provided got {len(dims)}"

        if n_bins is not None:
            assert isinstance(
                n_bins, Sequence
            ), f"n_bins should be a sequence if dims is a sequence, got {type(n_bins)}"
            assert (
                len(n_bins) == 2
            ), f"Number of bins should be provided for 2 dimensions, got {len(n_bins)}"
            assert (
                n_bins[0] > 0 and n_bins[1] > 0
            ), f"n_bins should all be greater than 0, got {n_bins}"
            w1, w2 = dims[0] / n_bins[0], dims[1] / n_bins[1]
            assert w1 == w2, (
                "Dimensions of cells should be equal in "
                f"both dimensions got {w1} and {w2}"
            )
            n_bins = (n_bins[0], n_bins[1])
        else:
            assert (
                i_range is not None
            ), "If n_bins is not provided, i_range should be provided"
            n0 = dims[0] / i_range
            n1 = dims[1] / i_range
            assert isclose(round(n0), n0) and isclose(
                round(n1), n1
            ), "Dimensions should be a multiple of i_range"
            n_bins = (round(n0), round(n1))

        width = dims[0] / n_bins[0]
        dims = jnp.array(dims)

    else:
        if n_bins is not None:
            assert isinstance(
                n_bins, int
            ), "n_bins should be an integer value if dims is a float"
            assert n_bins > 0, f"n_bins should be greater than 0, got {n_bins}"
            n_bins = (n_bins, n_bins)
        else:
            n_bins = floor(dims / i_range)
            n_bins = (n_bins, n_bins)

        width = dims / n_bins[0]
        dims = jnp.array([dims, dims])

    return n_bins, width, dims


def _check_arguments(
    pos: chex.Array,
    pos_b: Optional[chex.Array],
    agent_a: chex.ArrayTree,
    agent_b: chex.ArrayTree,
) -> None:
    assert (
        pos.ndim == 2 and pos.shape[1] == 2
    ), f"pos argument should be an array of 2d coordinates, got shape {pos.shape}"

    if pos_b is not None:
        assert pos.ndim == 2 and pos.shape[1] == 2, (
            "pos_b argument should be an array of "
            f"2d coordinates, got shape {pos.shape}"
        )
        n_b = pos_b.shape[:1]
    else:
        n_b = pos.shape[:1]

    if agent_a is not None:
        chex.assert_tree_has_only_ndarrays(agent_a)
        chex.assert_tree_shape_prefix(agent_a, pos.shape[:1])

    if agent_b is not None:
        chex.assert_tree_has_only_ndarrays(agent_b)
        chex.assert_tree_shape_prefix(agent_b, n_b)


def spatial(
    f: Callable,
    *,
    reduction: Reduction,
    include_self: bool = False,
    topology: str = "moore",
    n_bins: Optional[int | Sequence[int]] = None,
    i_range: Optional[float] = None,
    dims: Union[float, Sequence[float]] = 1.0,
) -> Callable:
    """
    Apply a function between agents based on spatial proximity

    This transformation efficiently checks
    for spatial proximity between agents before applying
    the interaction function to pairs of agents
    below a threshold range of each other.

    .. warning::

       This implementation currently assumes a 2-dimensional
       space with continues boundary conditions (i.e. wrapped
       on a torus). The shape/dimensions of the space
       can be controlled with the `dims` parameter, by default
       it is a unit square region.

    .. note::

       Performance is dependent on the proximity threshold
       and density of agents. Worse case it will perform
       :math:`N^2` checks if all agents are in the same
       neighbourhood.

    Examples
    --------

    This example subdivides the space into 2 cells along
    each dimension (i.e. there are 4 cells total),
    then sums contributions from neighbours, excluding the
    active agent

    .. testsetup:: spatial

       import esquilax
       import jax
       import jax.numpy as jnp
       from functools import partial

    .. testcode:: spatial

       def foo(p, a, b):
           return p + a + b

       x = jnp.array([[0.1, 0.1], [0.6, 0.6], [0.7, 0.7]])
       a = jnp.arange(3)

       result = esquilax.transforms.spatial(
           foo,
           i_range=0.5,
           reduction=esquilax.reductions.add(dtype=int),
           include_self=False,
       )(
           2, a, a, pos=x
       )
       # result = [0, 5, 5]

    .. doctest:: spatial
       :hide:

       >>> result
       Array([0, 5, 5], dtype=int32)

    so in this case agent ``0`` does not have any neighbours in
    its cell, but agents ``1`` and ``2`` observe each other.

    The transform can also be used as a decorator using
    :py:meth:`functools.partial`. Arguments and return values
    can be PyTrees or multidimensional arrays. Arguments can
    also be ``None`` if not used

    .. testcode:: spatial

       tuple_reduce = esquilax.reductions.Reduction(
           fn=(jnp.add, jnp.add), id=(0, 0),
       )

       @partial(
           esquilax.transforms.spatial,
           i_range=0.5,
           reduction=tuple_reduce,
           include_self=False,
           topology="same-cell",
       )
       def foo(p, _, b):
           return p + b[0],  p + b[1]

       x = jnp.array([[0.1, 0.1], [0.6, 0.6], [0.7, 0.7]])
       a = jnp.arange(6).reshape(3, 2)

       result = foo(2, None, a, pos=x)
       # result = ([0, 6, 4], [0, 7, 5])

    .. doctest:: spatial
       :hide:

       >>> result
       (Array([0, 6, 4], dtype=int32), Array([0, 7, 5], dtype=int32))

    You can also pass different agent types (i.e. two sets
    of agents with different positions) using the ``pos_b``
    keyword argument

    .. testcode:: spatial

       @partial(
           esquilax.transforms.spatial,
           i_range=0.5,
           reduction=esquilax.reductions.add(dtype=int),
           topology="moore",
       )
       def foo(params, a, b):
           return params + a + b

       # A consists of 3 agents, and b 2 agents
       xa = jnp.array([[0.1, 0.1], [0.1, 0.7], [0.75, 0.2]])
       xb = jnp.array([[0.7, 0.1], [0.75, 0.75]])
       vals_a = jnp.arange(3)
       vals_b = jnp.arange(1, 3)

       result = foo(2, vals_a, vals_b, pos=xa, pos_b=xb)
       # result = [22, 10, 17]

    .. doctest:: spatial
       :hide:

       >>> result
       Array([22, 10, 17], dtype=int32)

    Random keys can be passed to the wrapped function
    by including the ``key`` keyword argument

    .. testcode:: spatial

       def foo(_p, _a, _b, *, key):
           return jax.random.choice(key, 100, ())

       k = jax.random.PRNGKey(101)
       result = esquilax.transforms.spatial(
           foo,
           i_range=0.5,
           reduction=esquilax.reductions.add(dtype=int),
           include_self=False,
       )(
           None, None, None, pos=x, key=k
       )

    Parameters
    ----------
    f
        Interaction to apply to in-proximity pairs, should
        have the signature

        .. code-block:: python

           def f(
               params: Any,
               a: Any,
               b: Any,
               **static_kwargs,
           ):
               ...
               return x

        where

        - ``params``: Parameters broadcast over all interactions
        - ``a``: Start agent in the interaction
        - ``b``: End agent in the interaction
        - ``**static_kwargs``: Any arguments required at compile
          time by JAX can be passed as keyword arguments.

        JAX random keys can be passed to the function by including
        the ``key`` keyword argument.
    reduction
        Binary monoidal reduction function.
    include_self
        if ``True`` each agent will include itself in the
        gathered values.
    topology
        Topology of cells, default ``"moore"``. Since cells
        interact with their neighbours, topologies with
        fewer neighbours can increase performance at the
        cost of fidelity. Should be one of ``"same-cell"``,
        ``"von-neumann"`` or ``"moore"``.
    i_range
        Range at which agents interact. Can be ommited, in which
        case the width of a cell is used as the interaction range
        (derived from ``n_bins``), but this can be increased/decreased
        using ``i_range`` dependent on the use-case.
    n_bins
        Optional number of bins each dimension is subdivided
        into. Assumes that each dimension contains the
        same number of cells. Each cell can only interact
        with adjacent cells, so this value also consequently
        also controls the number of interactions. If not provided
        the minimum number of bins if derived from ``i_range``. For a square
        space ``n_bins`` can be a single intiger, or a pair of integers
        for the number of bins along each cell. The number of cells for
        each dimension should result in square cells.
    dims
        Dimensions of the space, either a float edge length for a
        square space, or a pait (tuple or list) of dimension.
        Default value is a square space of size 1.0.
    """

    n_bins, width, dims = _process_parameters(i_range, dims, n_bins)
    i_range = width if i_range is None else i_range
    i_range = i_range**2
    offsets = utils.space.get_neighbours_offsets(topology)
    keyword_args = utils.functions.get_keyword_args(f)
    has_key, keyword_args = utils.functions.has_key_keyword(keyword_args)

    @partial(jax.jit, static_argnames=keyword_args)
    def _spatial(
        params: Any,
        agents_a: Any,
        agents_b: Any,
        *,
        pos: chex.Array,
        pos_b: Optional[chex.Array] = None,
        key: Optional[chex.PRNGKey] = None,
        **static_kwargs,
    ) -> Any:
        _check_arguments(pos, pos_b, agents_a, agents_b)
        utils.functions.check_key(has_key, key)

        same_types = pos_b is None

        (
            co_ords_a,
            sort_idxs_a,
            bins_a,
            sorted_idxs_a,
            sorted_pos_a,
            sorted_agents_a,
        ) = _sort_agents(n_bins, width, pos, agents_a)

        if same_types:
            bins_b = bins_a
            sorted_pos_b = sorted_pos_a
            sorted_agents_b = jax.tree_util.tree_map(lambda y: y[sort_idxs_a], agents_b)
        else:
            _, _, bins_b, _, sorted_pos_b, sorted_agents_b = _sort_agents(
                n_bins, width, pos_b, agents_b
            )

        def cell(k: Optional[chex.PRNGKey], i: int, bin_range: chex.Array) -> Any:
            agent_a = jax.tree_util.tree_map(lambda y: y[i], sorted_agents_a)
            pos_a = sorted_pos_a[i]

            def interact(
                j: int, _k: Optional[chex.PRNGKey], _r: Any
            ) -> Tuple[Optional[chex.PRNGKey], Any]:
                agent_b = jax.tree_util.tree_map(lambda z: z[j], sorted_agents_b)
                if has_key:
                    _k, fk = jax.random.split(_k, 2)
                    r = f(params, agent_a, agent_b, key=fk, **static_kwargs)
                else:
                    r = f(params, agent_a, agent_b, **static_kwargs)
                r = jax.tree_util.tree_map(
                    lambda red, a, b: red(a, b), reduction.fn, _r, r
                )
                return _k, r

            def inner(
                j: int, carry: Tuple[Optional[chex.PRNGKey], Any]
            ) -> Tuple[Optional[chex.PRNGKey], Any]:
                _k, _r = carry
                _pos_b = sorted_pos_b[j]
                d = utils.space.shortest_distance(
                    pos_a, _pos_b, length=dims, norm=False
                )
                return jax.lax.cond(
                    d <= i_range, interact, lambda _, _x, _z: (_x, _z), j, _k, _r
                )

            if (not same_types) or include_self:
                _, _results = jax.lax.fori_loop(
                    bin_range[0], bin_range[1], inner, (k, reduction.id)
                )
            else:
                k, _results = jax.lax.fori_loop(
                    bin_range[0],
                    jnp.minimum(i, bin_range[1]),
                    inner,
                    (k, reduction.id),
                )
                _, _results = jax.lax.fori_loop(
                    jnp.maximum(i + 1, bin_range[0]),
                    bin_range[1],
                    inner,
                    (k, _results),
                )

            return _results

        def agent_reduce(
            k: Optional[chex.PRNGKey], i: chex.Numeric, co_ords: chex.Array
        ) -> chex.ArrayTree:
            nbs = utils.space.neighbour_indices(co_ords, offsets, n_bins)
            nb_bins = bins_b[nbs]

            if has_key:
                _keys = jax.random.split(k, nbs.shape[0])
                _results = jax.vmap(cell, in_axes=(0, None, 0))(_keys, i, nb_bins)
            else:
                _results = jax.vmap(cell, in_axes=(None, None, 0))(None, i, nb_bins)

            def red(a, _, c):
                return jnp.frompyfunc(c, 2, 1).reduce(a)

            return jax.tree_util.tree_map(red, _results, reduction.id, reduction.fn)

        n_agents = pos.shape[0]

        if has_key:
            keys = jax.random.split(key, n_agents)
            results = jax.vmap(agent_reduce, in_axes=(0, 0, 0))(
                keys, jnp.arange(n_agents), co_ords_a
            )
        else:
            results = jax.vmap(agent_reduce, in_axes=(None, 0, 0))(
                None, jnp.arange(n_agents), co_ords_a
            )

        inv_sort = jnp.argsort(sort_idxs_a)
        results = jax.tree_util.tree_map(lambda y: y[inv_sort], results)

        return results

    return _spatial


def nearest_neighbour(
    f: Callable,
    *,
    default: Default,
    topology: str = "moore",
    n_bins: Optional[int | Sequence[int]] = None,
    i_range: Optional[float] = None,
    dims: Union[float, Sequence[float]] = 1.0,
) -> Callable:
    """
    Apply a function between an agent and its closest neighbour

    This transformation locates the nearest neighbour
    to each agent and then applies the update function
    to this pair. If no nearest neighbour then the
    default value is returned.

    .. warning::

       This implementation currently assumes a 2-dimensional
       space with continues boundary conditions (i.e. wrapped
       on a torus). The shape/dimensions of the space
       can be controlled with the `dims` parameter, by default
       it is a unit square region.

    .. note::

       Performance is dependent on the proximity threshold
       and density of agents. Worse case it will perform
       :math:`N^2` checks if all agents are in the same
       neighbourhood.

    Examples
    --------

    This example subdivides the space into 2 cells along
    each dimension (i.e. there are 4 cells total) then
    locate closest neighbours

    .. testsetup:: spatial

       import esquilax
       import jax
       import jax.numpy as jnp
       from functools import partial

    .. testcode:: spatial

       def foo(p, a, b):
           return p + a + b

       x = jnp.array([[0.4, 0.4], [0.6, 0.6], [0.7, 0.7]])
       a = jnp.arange(3)

       result = esquilax.transforms.nearest_neighbour(
           foo,
           i_range=0.5,
           default=-1,
           topology="moore"
       )(
           2, a, a, pos=x
       )
       # result = [3, 5, 5]

    .. doctest:: spatial
       :hide:

       >>> result
       Array([3, 5, 5], dtype=int32)

    so in this case agents ``1`` and ``2`` are nearest neighbours
    and ``0`` is closest to ``1``.

    The transform can be used as a decorator using
    :py:meth:`functools.partial`. Arguments and return values
    can be PyTrees or multidimensional arrays. Arguments can also
    be ``None`` if not used

    .. testcode:: spatial

       @partial(
           esquilax.transforms.nearest_neighbour,
           i_range=0.5,
           default=(-1, -2),
           topology="moore",
       )
       def foo(p, _, b):
           return p + b[0],  p + b[1]

       x = jnp.array([[0.1, 0.1], [0.6, 0.6], [0.7, 0.7]])
       a = jnp.arange(6).reshape(3, 2)

       result = foo(2, None, a, pos=x)
       # result = ([-1, 6, 4], [-2, 7, 5])

    .. doctest:: spatial
       :hide:

       >>> result
       (Array([-1,  6,  4], dtype=int32), Array([-2,  7,  5], dtype=int32))

    You can also pass different agent types (i.e. two sets
    of agents with different positions) using the ``pos_b``
    keyword argument

    .. testcode:: spatial

       @partial(
           esquilax.transforms.nearest_neighbour,
           i_range=0.5,
           default=-1,
           topology="moore",
       )
       def foo(params, a, b):
           return params + a + b

       # A consists of a single agents, and b 2 agents
       xa = jnp.array([[0.1, 0.1]])
       xb = jnp.array([[0.2, 0.2], [0.75, 0.75]])
       vals_a = jnp.array([1])
       vals_b = jnp.array([2, 12])

       result = foo(2, vals_a, vals_b, pos=xa, pos_b=xb)
       # result = [5]

    .. doctest:: spatial
       :hide:

       >>> result
       Array([5], dtype=int32)

    Random keys can be passed to the wrapped function by
    using the ``key`` keyword argument

    .. testcode:: spatial

       def foo(p, a, b, *, key):
           return jax.random.choice(key, 100, ())

       k = jax.random.PRNGKey(101)
       result = esquilax.transforms.nearest_neighbour(
           foo,
           i_range=0.5,
           default=-1,
           topology="moore"
       )(
           None, None, None, pos=x, key=k
       )

    Parameters
    ----------
    f
        Interaction to apply to in-proximity pairs, should
        have the signature

        .. code-block:: python

           def f(
               params: Any,
               a: Any,
               b: Any,
               **static_kwargs,
           ):
               ...
               return x

        where

        - ``k``: Is a JAX random key
        - ``params``: Parameters broadcast over all interactions
        - ``a``: Start agent in the interaction
        - ``b``: End agent in the interaction
        - ``**static_kwargs``: Any arguments required at compile
          time by JAX can be passed as keyword arguments.

        Random keys can be passed to the function by including
        the ``key`` keyword argument.
    default
        Default value(s) returned if no-neighbours are in
        range of an agent.
    topology
        Topology of cells, default ``"moore"``. Since cells
        interact with their neighbours, topologies with
        fewer neighbours can increase performance at the
        cost of fidelity. Should be one of ``"same-cell"``,
        ``"von-neumann"`` or ``"moore"``.
    i_range
        Range at which agents interact. Can be ommited, in which
        case the width of a cell is used as the interaction range
        (derived from ``n_bins``), but this can be increased/decreased
        using ``i_range`` dependent on the use-case.
    n_bins
        Optional number of bins each dimension is subdivided
        into. Assumes that each dimension contains the
        same number of cells. Each cell can only interact
        with adjacent cells, so this value also consequently
        also controls the number of interactions. If not provided
        the minimum number of bins if derived from ``i_range``.
    dims
        Dimensions of the space, either a float edge length for a
        square space, or a pait (tuple or list) of dimension.
        Default value is a square space of size 1.0.
    """
    n_bins, width, dims = _process_parameters(i_range, dims, n_bins)
    i_range = width if i_range is None else i_range
    i_range = i_range**2

    offsets = utils.space.get_neighbours_offsets(topology)
    keyword_args = utils.functions.get_keyword_args(f)
    has_key, keyword_args = utils.functions.has_key_keyword(keyword_args)

    @partial(jax.jit, static_argnames=keyword_args)
    def _nearest_neighbour(
        params: Any,
        agents_a: Any,
        agents_b: Any,
        *,
        pos: chex.Array,
        pos_b: Optional[chex.Array] = None,
        key: Optional[chex.PRNGKey] = None,
        **static_kwargs,
    ) -> Any:
        _check_arguments(pos, pos_b, agents_a, agents_b)
        utils.functions.check_key(has_key, key)

        same_types = pos_b is None

        (
            co_ords_a,
            sort_idxs_a,
            bins_a,
            sorted_idxs_a,
            sorted_pos_a,
            sorted_agents_a,
        ) = _sort_agents(n_bins, width, pos, agents_a)

        if same_types:
            bins_b = bins_a
            sorted_pos_b = sorted_pos_a
            sorted_agents_b = jax.tree_util.tree_map(lambda y: y[sort_idxs_a], agents_b)
        else:
            _, _, bins_b, _, sorted_pos_b, sorted_agents_b = _sort_agents(
                n_bins, width, pos_b, agents_b
            )

        def cell(i: int, bin_range: chex.Array) -> Tuple[int, float]:
            pos_a = sorted_pos_a[i]

            def inner(j: int, carry: Tuple[int, float]) -> Tuple[int, float]:
                _best_idx, _best_d = carry
                _pos_b = sorted_pos_b[j]
                _d = utils.space.shortest_distance(
                    pos_a, _pos_b, length=dims, norm=False
                )
                return jax.lax.cond(
                    jnp.logical_and(_d <= i_range, _d < _best_d),
                    lambda: (j, _d),
                    lambda: (_best_idx, _best_d),
                )

            if not same_types:
                best_idx, best_d = jax.lax.fori_loop(
                    bin_range[0], bin_range[1], inner, (-1, jnp.inf)
                )
            else:
                best_idx, best_d = jax.lax.fori_loop(
                    bin_range[0],
                    jnp.minimum(i, bin_range[1]),
                    inner,
                    (-1, jnp.inf),
                )
                best_idx, best_d = jax.lax.fori_loop(
                    jnp.maximum(i + 1, bin_range[0]),
                    bin_range[1],
                    inner,
                    (best_idx, best_d),
                )

            return best_idx, best_d

        def agent_reduce(i: chex.Numeric, co_ords: chex.Array) -> chex.Numeric:
            nbs = utils.space.neighbour_indices(co_ords, offsets, n_bins)
            nb_bins = bins_b[nbs]
            best_idx, best_d = jax.vmap(cell, in_axes=(None, 0))(i, nb_bins)
            min_idx = jnp.argmin(best_d)
            min_idx = best_idx[min_idx]
            return min_idx

        n_agents = pos.shape[0]
        nearest_idxs = jax.vmap(agent_reduce, in_axes=(0, 0))(
            jnp.arange(n_agents), co_ords_a
        )
        inv_sort = jnp.argsort(sort_idxs_a)
        nearest_idxs = nearest_idxs[inv_sort]

        def apply(
            k: Optional[chex.PRNGKey], a: chex.ArrayTree, idx_b: chex.Numeric
        ) -> chex.ArrayTree:
            b = jax.tree.map(lambda x: x.at[idx_b].get(), sorted_agents_b)
            if has_key:
                result = f(params, a, b, key=k, **static_kwargs)
            else:
                result = f(params, a, b, **static_kwargs)
            return result

        def check(
            k: Optional[chex.PRNGKey], a: chex.ArrayTree, idx_b: chex.Numeric
        ) -> chex.ArrayTree:
            return jax.lax.cond(
                idx_b < 0,
                lambda *_: default,
                apply,
                k,
                a,
                idx_b,
            )

        if has_key:
            keys = jax.random.split(key, n_agents)
            results = jax.vmap(check)(keys, agents_a, nearest_idxs)
        else:
            results = jax.vmap(check, in_axes=(None, 0, 0))(
                None, agents_a, nearest_idxs
            )

        return results

    return _nearest_neighbour



"""
`Evosax`_ policy training and testing functionality

Basic training and testing loops of neuro-evolution
strategies. Also supports training multiple strategies
within the same loop.

.. _Evosax: https://github.com/RobertTLange/evosax
"""
from __future__ import annotations

from functools import partial
from typing import Collection, Optional, Tuple

import chex
import evosax
import jax
import jax.numpy as jnp
import jax_tqdm

from esquilax.batch_runner import batch_sim_runner
from esquilax.sim import Sim
from esquilax.typing import TSimParams, TypedPyTree

from . import tree_utils
from .strategy import Strategy
from .types import TrainingData


@partial(
    jax.jit,
    static_argnames=(
        "strategies",
        "env",
        "n_generations",
        "n_env",
        "n_steps",
        "map_population",
        "show_progress",
    ),
)
def train(
    strategies: TypedPyTree[Strategy],
    env: Sim,
    n_generations: int,
    n_env: int,
    n_steps: int,
    map_population: bool,
    key: chex.PRNGKey,
    evo_params: TypedPyTree[evosax.EvoParams],
    evo_states: TypedPyTree[evosax.EvoState],
    show_progress: bool = True,
    env_params: Optional[TSimParams] = None,
) -> Tuple[Collection[evosax.EvoState], chex.ArrayTree]:
    """
    Train a policy/policies using neuro-evolution

    Training loop that optimises a policy or multiple
    policies evaluated against an Esquilax simulation
    training environment.

    Agent policies/behaviours are sampled and updated
    using Evosax, and can either be tested as
    a shard policy (i.e. all agents share a policy
    from the population), or each agent is controlled
    by an individual member of the population.

    The training loop broadly follows these steps:

    - Sample a new population from the strategy
    - Initialise a simulation(s) and run with
      the population/parameters
    - Collect rewards over the simulation
    - Aggregate and rescale rewards
    - Update the strategy state from the
      rewards/fitness

    .. note::

       Rewards are summed over the simulation steps, i.e.
       the total rewards for each agent (or policy)
       are used to measure fitness during training.

    Parameters
    ----------
    strategies
        PyTree of Strategy classes. This could be either a
        single Strategy, or a container/struct of strategies.
    env
        Esquilax simulation training environment

        .. warning::

           The step function of the environment should
           include the population/parameters as
           a keyword argument `agent_params`, and
           return a `TrainingData` class, for example:

           .. code-block:: python

              def step(self, i, k, params, state, *, agent_params):
                  ...
                  # Rewards are returned as part of the TrainingData
                  return (
                      state,
                      TrainingData(rewards=rewards, records=records)
                  )

           The current population and rewards will then be
           passed and handled automatically by the training
           loop.

           If multiple strategies are being used, then the returned
           training data should have the same structure, e.g.:

           .. code-block:: python

              # Multiple strategies as a dictionary
              strategies = {"a": ..., "b": ...}

              def step(self, i, k, params, state, *, agent_params):
                  ...
                  # Should return dict matching the strategy structure
                  return (
                      state,
                      dict(
                        a = TrainingData(...),
                        b = TrainingData(...),
                      )
                  )

    n_generations
        Number of training generations
    n_env
        Number of Monte-Carlo samples to test each population
        or parameter samples across.
    n_steps
        Number of steps to run each simulation
    map_population
        If ``True`` each member of the population is
        evaluated in a separate environment (i.e. the
        individual parameters are shared by agents). If
        ``False`` the whole population is passed
        to the environment as part of the simulation
        state.
    key
        JAX random key
    evo_params
        Evosax strategy parameters
    evo_states
        Evosax strategy state
    show_progress
        If ``True`` a progress bar will be displayed
        showing training progress. Default value is ``True``.
    env_params
        Optional environment parameters, if not provided the
        default environment parameters will be used.

    Returns
    -------
    tuple[evosax.EvoState, chex.ArrayTree]
        Trained strategy state and record of rewards
        gathered over the course of training. If multiple
        strategies were provided, then the output will
        have the same structure.
    """

    env_params = env.default_params() if env_params is None else env_params

    def generation(carry, _):
        _k, _evo_states = carry

        _k, k1, k2 = jax.random.split(_k, 3)

        _evo_states, _population, _population_shaped = tree_utils.tree_ask(
            k1, strategies, _evo_states, evo_params
        )

        def inner(_pop) -> TrainingData:
            return batch_sim_runner(
                env,
                n_env,
                n_steps,
                k2,
                show_progress=False,
                params=env_params,
                agent_params=_pop,
            )

        if map_population:
            _training_data = jax.vmap(inner)(_population_shaped)
            _total_rewards = jax.tree.map(
                lambda t: jnp.sum(t.rewards, axis=(0, 2, 3)),
                _training_data,
                is_leaf=lambda x: isinstance(x, TrainingData),
            )
        else:
            _training_data = inner(_population_shaped)
            _total_rewards = jax.tree.map(
                lambda t: jnp.sum(t.rewards, axis=(0, 1)),
                _training_data,
                is_leaf=lambda x: isinstance(x, TrainingData),
            )

        _fitness, _evo_states = tree_utils.tree_tell(
            strategies, _population, _total_rewards, _evo_states, evo_params
        )

        return (_k, _evo_states), _total_rewards

    if show_progress:
        generation = jax_tqdm.scan_tqdm(n_generations, desc="Generation")(generation)

    (_, evo_states), rewards = jax.lax.scan(
        generation,
        (key, evo_states),
        jnp.arange(n_generations),
    )

    return evo_states, rewards


@partial(
    jax.jit,
    static_argnames=(
        "env",
        "n_env",
        "n_steps",
        "map_population",
        "show_progress",
    ),
)
def test(
    population_shaped: chex.ArrayTree,
    env: Sim,
    n_env: int,
    n_steps: int,
    map_population: bool,
    key: chex.PRNGKey,
    show_progress: bool = True,
    env_params: Optional[TSimParams] = None,
) -> TrainingData:
    """
    Test population performance and gather telemetry

    Test a population against a simulation environment
    gathering rewards and recorded simulation state
    over the course of the simulation

    Parameters
    ----------
    population_shaped
        Reshaped population parameters for use
        by simulation agents.
    env
        Esquilax simulation training environment
    n_env
        Number of Monte-Carlo samples to test each population
        or parameter samples across.
    n_steps
        Number of simulation steps.
    map_population
        If ``True`` each member of the population is
        evaluated in a separate environment (i.e. the
        individual parameters are shared by agents). If
        ``False`` the whole population is passed
        to the environment as part of the simulation
        state.
    key
        JAX random key
    show_progress
        If ``True`` a progress bar will be displayed
        showing simulation progress. Default value is ``True``.
    env_params
        Optional environment parameters, if not provided the
        default environment parameters will be used.

    Returns
    -------
    esquilax.ml.evo.TrainingData
        Data collected over the course of the simulation.
    """
    k1, k2 = jax.random.split(key)

    env_params = env.default_params() if env_params is None else env_params

    def inner(_pop):
        return batch_sim_runner(
            env,
            n_env,
            n_steps,
            k2,
            show_progress=show_progress,
            params=env_params,
            agent_params=_pop,
        )

    if map_population:
        records = jax.vmap(inner)(population_shaped)
    else:
        records = inner(population_shaped)

    return records



"""
RL related data classes
"""
import chex


@chex.dataclass
class Trajectory:
    """
    Observation-action environment trajectories

    Parameters
    ----------
    obs
        Agent observations
    actions
        Agent actions
    action_values
        Any additional values associated with
        the action
    rewards
        Agent rewards
    done
        Terminal flag
    """

    obs: chex.ArrayTree
    actions: chex.ArrayTree
    action_values: chex.ArrayTree
    rewards: chex.ArrayTree
    done: chex.ArrayTree



"""
Reinforcement learning utilities and training functionality

.. note::

   This module requires additional dependencies that can be installed
   using the installation extra ``rl``, for example with pip using

   .. code::

      pip install esquilax[rl]
"""
try:
    import flax  # noqa: F401
except ImportError as e:
    raise ImportError(
        "This functionality requires additional dependencies."
        "They can be installed with the extra `pip install esquilax[rl]`"
    ) from e
from .agent import Agent
from .agent_state import AgentState, BatchAgentState
from .environment import Environment
from .training import test, train, train_and_test
from .types import Trajectory



"""
RL training and testing functionality
"""
from functools import partial
from typing import Callable, Tuple, Union

import chex
import jax
import jax.numpy as jnp
import jax_tqdm

from esquilax.typing import TEnvParams, TEnvState, TypedPyTree

from . import tree_utils
from .agent import Agent
from .agent_state import AgentState, BatchAgentState
from .environment import Environment
from .types import Trajectory


def step(
    agents: TypedPyTree[Agent],
    env_params: TEnvParams,
    env: Environment,
    greedy: bool = False,
) -> Callable:
    """
    Single update of the training environment

    Returns a step function intended for use inside
    a JAX scan. The step samples agent actions
    from the current state and updates the environment.

    Parameters
    ----------
    agents
        PyTree of RL agents
    env_params
        Environment parameters
    env
        Training environment
    greedy
        Flag to be passed to RL agent, indicating sampling
        actions from a greedy policy

    Returns
    -------
    typing.Callable
        Step function for use in :py:meth:`jax.lax.scan`.
        Carries random keys, environment state and observations.
        Then returns transitions/trajectories and environment
        state to be recorded.
    """

    def inner_step(carry, _) -> Tuple[Tuple, Tuple[Trajectory, TEnvState]]:
        _k, _env_state, _obs, _agent_states = carry
        _k, _k_act, _k_step = jax.random.split(_k, 3)
        _actions, _action_values = tree_utils.sample_actions(
            agents,
            _k_act,
            _agent_states,
            _obs,
            greedy=greedy,
        )
        _new_obs, _new_env_state, _rewards, _done = env.step(
            _k_step, env_params, _env_state, _actions
        )
        _trajectories = jax.tree.map(
            lambda _, *x: Trajectory(
                obs=x[0],
                actions=x[1],
                action_values=x[2],
                rewards=x[3],
                done=x[4],
            ),
            agents,
            _obs,
            _actions,
            _action_values,
            _rewards,
            _done,
            is_leaf=lambda x: isinstance(x, Agent),
        )
        return (
            (_k, _new_env_state, _new_obs, _agent_states),
            (_trajectories, _env_state),
        )

    return inner_step


def generate_samples(
    agents: TypedPyTree[Agent],
    env_params: TEnvParams,
    env: Environment,
    n_env_steps: int,
    key: chex.PRNGKey,
    agent_states: TypedPyTree[Union[AgentState, BatchAgentState]],
    greedy: bool = False,
    show_progress: bool = False,
    pbar_id: int = 0,
) -> Tuple[Trajectory, TEnvState]:
    """
    Run the environment forward generating trajectory and state records

    Run the simulation environment and collecting trajectory and
    environment state records.

    Parameters
    ----------
    agents
        PyTree of RL agents.
    env_params
        Environment parameters.
    env
        Training environment.
    n_env_steps
        Number of steps to run environment.
    key
        JAX random key
    agent_states
        PyTree of RL agent states
    greedy
        Flag to be passed to RL agent, indicating sampling
        actions from a greedy policy.
    show_progress
        If ``True`` a progress bar will show execution progress.
    pbar_id
        Optional progress bar index. Can be used to print
        to multiple progress bars.

    Returns
    -------
    tuple[esquilax.ml.rl.Trajectory, esquilax.typing.TEnvState]
        Environment observation-action trajectories and
        recorded environment states. Trajectories have a shape
        ``[n-steps, n-agents]``.
    """
    step_fun = step(agents, env_params, env, greedy=greedy)

    k_reset, k_run = jax.random.split(key, 2)
    obs, env_state = env.reset(k_reset, env_params)
    init = (k_run, env_state, obs, agent_states)

    if show_progress:
        step_fun = jax_tqdm.scan_tqdm(n_env_steps, desc="Step")(step_fun)
        init = jax_tqdm.PBar(id=pbar_id, carry=init)

    _, (trajectories, env_states) = jax.lax.scan(
        step_fun, init, jnp.arange(n_env_steps)
    )
    return trajectories, env_states


def batch_generate_samples(
    agents: TypedPyTree[Agent],
    env_params: TEnvParams,
    env: Environment,
    n_env_steps: int,
    n_env: int,
    key: chex.PRNGKey,
    agent_states: TypedPyTree[Union[AgentState, BatchAgentState]],
    greedy: bool = False,
    show_progress: bool = False,
) -> Tuple[Trajectory, TEnvState]:
    """
    Sample trajectories across multiple environments

    Generate samples across multiple environments, across
    random seeds.

    Parameters
    ----------
    agents
        PyTree of RL agents.
    env_params
        Environment parameters.
    env
        Training environment.
    n_env_steps
        Number of steps to run environment.
    n_env
        Number of environments to execute.
    key
        JAX random key
    agent_states
        PyTree of RL agent states
    greedy
        Flag to be passed to RL agent, indicating sampling
        actions from a greedy policy.
    show_progress
        If ``True`` a progress bar will show execution progress.

    Returns
    -------
    tuple[esquilax.ml.rl.Trajectory, esquilax.typing.TEnvState]
        Environment observation-action trajectories and
        recorded environment states. Trajectories have a shape
        ``[n-env, n-steps, n-agents]``.
    """
    sampling_func = partial(
        generate_samples,
        agents,
        env_params,
        env,
        n_env_steps,
        greedy=greedy,
        show_progress=show_progress,
    )
    keys = jax.random.split(key, n_env)
    trajectories, env_states = jax.vmap(
        lambda k, a, i: sampling_func(k, a, pbar_id=i), in_axes=(0, None, 0)
    )(keys, agent_states, jnp.arange(n_env))
    return trajectories, env_states


def train(
    key: chex.PRNGKey,
    agents: TypedPyTree[Agent],
    agent_states: TypedPyTree[Union[AgentState, BatchAgentState]],
    env: Environment,
    env_params: TEnvParams,
    n_epochs: int,
    n_env: int,
    n_env_steps: int,
    show_progress: bool = True,
) -> Tuple[TypedPyTree[Agent], chex.ArrayTree, chex.ArrayTree]:
    """
    Train an RL-agent or agents with a given environment

    Parameters
    ----------
    key
        JAX random key.
    agents
        RL agent, or collection of agents functionality.
    agent_states
        Corresponding RL agent(s) states
    env
        Training environment/simulation. This should
        implement the :py:class:`esquilax.ml.rl.Environment`
        interface.
    env_params
        Environment parameters.
    n_epochs
        Number of training epochs.
    n_env
        Number of environments to train across per epoch.
    n_env_steps
        Number of steps to run in each environment per epoch.
    show_progress
        If ``True`` a training progress bar will be displayed.
        Default ``True``.

    Returns
    -------
    tuple
        Tuple containing:

        - Update agent or collection of agents
        - Training rewards
        - Additional training data
    """

    batch_sampling_func = partial(
        batch_generate_samples,
        agents,
        env_params,
        env,
        n_env_steps,
        n_env,
        greedy=False,
        show_progress=False,
    )

    def epoch(carry, _):
        _k, _agent_states = carry
        _k, _k_sample, _k_train = jax.random.split(_k, 3)
        _trajectories, _ = batch_sampling_func(_k_sample, _agent_states)
        _agent_states, _train_data = tree_utils.update_agents(
            agents, _k_train, _agent_states, _trajectories
        )

        return (_k, _agent_states), (
            jax.tree.map(
                lambda _, t: t.rewards,
                agents,
                _trajectories,
                is_leaf=lambda x: isinstance(x, Agent),
            ),
            _train_data,
        )

    if show_progress:
        epoch = jax_tqdm.scan_tqdm(n_epochs, desc="Epoch")(epoch)

    (_, agent_states), (rewards, train_data) = jax.lax.scan(
        epoch,
        (key, agent_states),
        jnp.arange(n_epochs),
    )

    return agent_states, rewards, train_data


def test(
    key: chex.PRNGKey,
    agents: TypedPyTree[Agent],
    agent_states: TypedPyTree[Union[AgentState, BatchAgentState]],
    env: Environment,
    env_params: TEnvParams,
    n_env: int,
    n_env_steps: int,
    show_progress: bool = True,
    return_trajectories: bool = False,
    greedy_actions: bool = False,
) -> Tuple[TEnvState, Union[Trajectory, chex.ArrayTree]]:
    """
    Test agent(s) performance

    Asses agents against a test environment,
    returning trajectory data gathered over training.

    Parameters
    ----------
    key
        JAX random key.
    agents
        RL agent, or collection of agents functionality.
    agent_states
        Corresponding RL agent(s) states
    env
        Training environment/simulation. This should
        implement a Gymnax Environment base class.
    env_params
        Environment parameters.
    n_env
        Number of environments to test across.
    n_env_steps
        Number of steps to run in each environment.
    show_progress
        If ``True`` a testing progress bar will be displayed.
        Default ``True``
    return_trajectories
        If ``True`` recorded trajectory data will be returned
        along with recorded state. If ``False`` the states and
        recorded rewards will be returned. Default ``False``.

        .. warning::

           Recording trajectories could result in a large amount
           of data (given it will record observations, actions etc.
           for each individual agent).

    greedy_actions
        Flag to indicate actions should be greedily sampled.

    Returns
    -------
    tuple[esquilax.typing.TEnvState, esquilax.ml.rl.Trajectory | chex.ArrayTree]
        Update trajectories gathered over testing.
    """
    sampling_func = partial(
        generate_samples,
        agents,
        env_params,
        env,
        n_env_steps,
        greedy=greedy_actions,
        show_progress=show_progress,
    )

    def sample_trajectories(_k, i, _agent_states):
        _trajectories, _states = sampling_func(_k, _agent_states, pbar_id=i)
        if return_trajectories:
            return _states, _trajectories
        else:
            return _states, jax.tree.map(
                lambda _, x: x.rewards,
                agents,
                _trajectories,
                is_leaf=lambda x: isinstance(x, Agent),
            )

    k_sample = jax.random.split(key, n_env)

    states, recorded = jax.vmap(sample_trajectories, in_axes=(0, 0, None))(
        k_sample, jnp.arange(n_env), agent_states
    )

    return states, recorded


def train_and_test(
    key: chex.PRNGKey,
    agents: TypedPyTree[Agent],
    agent_states: TypedPyTree[Union[AgentState, BatchAgentState]],
    env: Environment,
    env_params: TEnvParams,
    n_train_steps: int,
    test_every: int,
    n_train_env: int,
    n_test_env: int,
    n_env_steps: int,
    show_progress: bool = True,
    return_trajectories: bool = False,
    greedy_test_actions: bool = False,
) -> Tuple[
    TypedPyTree[Union[AgentState, BatchAgentState]],
    chex.Array,
    chex.ArrayTree,
    chex.ArrayTree,
    chex.ArrayTree,
]:
    """
    RL train and test loop

    Train an RL agent(s) and test at fixed intervals.
    Agent rewards and losses are tracked over the
    course of training. Rewards (or optionally state
    transitions) and the state of the environment are
    recorded during testing.

    It's assumed that collecting samples across multiple
    agents and environments so the agent training is
    performed each training step.

    Parameters
    ----------
    key
        JAX random key.
    agents
        PyTree of RL agent definitions.
    agent_states
        PyTree of RL agent states.
    env
        Training environment definition.
    env_params
        Environment parameters.
    n_train_steps
        Total number of training steps.
    test_every
        Number of steps between testing.
    n_train_env
        Number of environments to run during training.
    n_test_env
        Number of environments to run during testing.
    n_env_steps
        Number of environment updates per episode.
    show_progress
        If ``True`` training/testing progress will
        be displayed.
    return_trajectories
        If ``True`` during testing trajectory data will be recorded
        along with environment state. If ``False`` the state and
        rewards will be recorded. Default ``False``.

        .. warning::

           Recording trajectories could result in a large amount
           of data (given it will record observations, actions etc.
           for each individual agent at each step).

    greedy_test_actions
        If ``True`` greedy action sampling flag will be passed
        to the RL agent(s) during testing.

    Returns
    -------
    tuple
        Tuple containing

        - PyTree of trained agent states
        - Individual agent rewarded recorded over training
        - Loss data recorded during training
        - Env state history recorded during testing
        - Rewards or trajectories recorded during testing
    """
    assert (
        n_train_steps % test_every == 0
    ), "n_train_steps should be a multiple of test_every"

    n_steps = n_train_steps // test_every

    def train_test_step(carry, _):
        _key, _agent_states = carry

        _key, k_train, k_test = jax.random.split(_key, 3)
        _agent_states, _train_rewards, _train_losses = train(
            k_train,
            agents,
            _agent_states,
            env,
            env_params,
            test_every,
            n_train_env,
            n_env_steps,
            show_progress=show_progress,
        )
        _env_state_records, _test_rewards = test(
            k_test,
            agents,
            _agent_states,
            env,
            env_params,
            n_test_env,
            n_env_steps,
            show_progress=show_progress,
            return_trajectories=return_trajectories,
            greedy_actions=greedy_test_actions,
        )
        return (_key, _agent_states), (
            _train_rewards,
            _train_losses,
            _env_state_records,
            _test_rewards,
        )

    (_, trained_agent_states), (
        train_rewards,
        train_losses,
        env_state_records,
        test_rewards,
    ) = jax.lax.scan(train_test_step, (key, agent_states), None, length=n_steps)

    def reshape_train_data(x):
        s0 = x.shape[0] * x.shape[1]
        return jnp.reshape(x, (s0,) + x.shape[2:])

    train_rewards = reshape_train_data(train_rewards)
    train_losses = jax.tree.map(reshape_train_data, train_losses)

    return (
        trained_agent_states,
        train_rewards,
        train_losses,
        env_state_records,
        test_rewards,
    )



"""
Abstract class wrapping common simulation functionality
"""
from functools import partial
from typing import Any, Generic, Optional, Tuple

import chex
import jax

from esquilax.runner import sim_runner

from .typing import TSimParams, TSimState


class Sim(Generic[TSimParams, TSimState]):
    """
    Base class wrapping simulation functionality for batch execution and ML use-cases

    Simulation environment base-class with methods used by training functions.
    For use inside JIT compiled training loops, static simulation
    parameters can be assigned as attributes of the derived class.
    """

    @partial(jax.jit, static_argnums=(0,))
    def default_params(self) -> TSimParams:
        """
        Return default simulation parameters

        Returns
        -------
        esquilax.typing.TSimParams
            Simulation parameters
        """
        raise NotImplementedError

    @partial(jax.jit, static_argnums=(0,))
    def initial_state(self, key: chex.PRNGKey, params: TSimParams) -> TSimState:
        """
        Initialise the initial state of the simulation

        Parameters
        ----------
        key
            JAX random key.
        params
            Simulation parameters

        Returns
        -------
        esquilax.typing.TSimState
            The initial state of the environment.
        """
        raise NotImplementedError

    @partial(jax.jit, static_argnums=(0,))
    def step(
        self,
        i: int,
        key: chex.PRNGKey,
        params: TSimParams,
        state: TSimState,
        **kwargs: Any
    ) -> Tuple[TSimState, chex.ArrayTree]:
        """
        A single step/update of the environment

        The step function should return a tuple containing the updated
        simulation state, and any data to be recorded each step
        (see :py:meth:`esquilax.runner.sim_runner` for more details).
        For example:

        .. code-block:: python

           class Sim(SimEnv):
               def step(self, i, k, params, state):
                   ...
                   return new_state, records

        Any static arguments required by the simulation can
        be accessed from the ``self`` argument of the method.

        Parameters
        ----------
        i
            Current step number
        key
            JAX random key
        params
            Simulation time-independent parameters
        state
            Simulation state
        **kwargs
            Any additional keyword arguments.

        Returns
        -------
        tuple[esquilax.typing.TSimState, chex.ArrayTree]
            Tuple containing the updated simulation state, and
            any data to be recorded over the course of the simulation.
        """
        raise NotImplementedError

    @partial(jax.jit, static_argnums=(0, 1), static_argnames=("show_progress",))
    def run(
        self,
        n_steps: int,
        key: chex.PRNGKey,
        params: TSimParams,
        initial_state: TSimState,
        show_progress: bool = True,
        pbar_id: int = 0,
        **step_kwargs: Any
    ) -> Tuple[TSimState, chex.ArrayTree, chex.PRNGKey]:
        """
        Convenience function to run the simulation for a fixed number of steps

        Parameters
        ----------
        n_steps
            Number of simulation steps
        key
            JAX random key
        params
            Simulation time-independent parameters
        initial_state
            Initial state of the simulation
        show_progress
            If ``True`` a progress bar will be displayed.
            Default ``True``
        pbar_id
            Optional progress bar index, can be used to print
            multiple progress bars.
        **step_kwargs
            Any additional keyword arguments passed to the
            step function. Arguments are static over the
            course of the simulation.

        Returns
        -------
        tuple[esquilax.typing.TSimState, chex.ArrayTree, jax,random.PRNGKey]
            Tuple containing

            - The final state of the simulation
            - Tree of recorded data
            - Updated JAX random key
        """
        final_state, records, k = sim_runner(
            partial(self.step, **step_kwargs),
            params,
            initial_state,
            n_steps,
            key,
            show_progress=show_progress,
            pbar_id=pbar_id,
        )

        return final_state, records, k

    @partial(jax.jit, static_argnums=(0, 1), static_argnames=("show_progress",))
    def init_and_run(
        self,
        n_steps: int,
        key: chex.PRNGKey,
        show_progress: bool = True,
        params: Optional[TSimParams] = None,
        pbar_id: int = 0,
        **step_kwargs
    ) -> Tuple[TSimState, chex.ArrayTree]:
        """
        Convenience function to initialise and run the simulation

        Parameters
        ----------
        n_steps
            Number of simulation steps to run
        key
            JAX random key
        show_progress
            If ``True`` a progress bar will be displayed.
            Default ``True``
        params
            Optional simulation parameters, if not provided
            default sim parameters will be used.
        pbar_id
            Optional progress bar index, can be used to print
            multiple progress bars.
        **step_kwargs
            Any additional keyword arguments passed to the
            step function. Arguments are static over the
            course of the simulation.

        Returns
        -------
        tuple[esquilax.typing.TSimState, chex.ArrayTree, jax,random.PRNGKey]
            Tuple containing

            - The final state of the simulation
            - Tree of recorded data
            - Updated JAX random key
        """
        k1, k2 = jax.random.split(key, 2)

        params = self.default_params() if params is None else params
        initial_state = self.initial_state(k1, params)
        final_state, records, _ = self.run(
            n_steps,
            k2,
            params,
            initial_state,
            show_progress=show_progress,
            pbar_id=pbar_id,
            **step_kwargs
        )

        return final_state, records



"""
Graph transformations

Transformations representing interactions between
nodes (and edges) on a graph.
"""
from functools import partial
from typing import Any, Callable, Optional

import chex
import jax
import jax.numpy as jnp

from esquilax import utils
from esquilax.reductions import Reduction
from esquilax.typing import Default


def _check_edges(edges: chex.ArrayTree, edge_idxs: chex.Array):
    assert edge_idxs.ndim == 2 and edge_idxs.shape[0] == 2, (
        "edge_idxs should be a 2d array of node " f"index pairs, got {edge_idxs.shape}"
    )
    chex.assert_tree_shape_prefix(edges, edge_idxs.shape[1:])


def edge_map(f: Callable) -> Callable:
    """
    Map a function over graph edges and related nodes

    Maps a function over a set of edges (with data) and
    data corresponding to the edge start and end nodes,

    .. warning::

       Edge indices and any associated data should be
       sorted using :py:meth:`esquilax.utils.sort_edges`

    Examples
    --------

    .. testsetup:: edge_map

       import esquilax
       import jax
       import jax.numpy as jnp

    .. testcode:: edge_map

       def f(params, start, end, edge):
           return params + start + end + edge

       edge_idxs = jnp.array([[0, 0, 1], [1, 2, 0]])
       edges = jnp.array([0, 1, 2])
       starts = jnp.array([0, 1, 2])
       ends = jnp.array([0, 1, 2])

       # Call transform with edge indexes
       result = esquilax.transforms.edge_map(
           f
       )(
           2, starts, ends, edges, edge_idxs=edge_idxs
       )
       # result = [3, 5, 5]

    .. testcode:: edge_map
       :hide:

       assert result.tolist() == [3, 5, 5]

    The transform can also be used as a decorator. Arguments
    can also be PyTrees or ``None`` if unused

    .. testcode:: edge_map

       @esquilax.transforms.edge_map
       def f(_params, _start, end, edge):
           return end[0] + end[1] + edge

       edge_idxs = jnp.array([[0, 0, 1], [1, 2, 0]])
       edges = jnp.array([0, 1, 2])
       ends = (jnp.array([0, 1, 2]), jnp.array([0, 1, 2]))

       # Call transform with edge indexes
       f(None, None, ends, edges, edge_idxs=edge_idxs)
       # [2, 5, 2]

    .. testcode:: edge_map
       :hide:

       z = f(None, None, ends, edges, edge_idxs=edge_idxs)
       assert z.tolist() == [2, 5, 2]

    JAX random keys can be passed to the wrapped function
    by including the ``key`` keyword argumnt

    .. testcode:: edge_map

       @esquilax.transforms.edge_map
       def f(_params, _start, _end, _edge, *, key):
           # Sample a random integer for each edge
           return jax.random.choice(key, 100, ())

       k = jax.random.PRNGKey(101)
       result = f(None, None, None, None, edge_idxs=edge_idxs, key=k)

    Parameters
    ----------
    f
        Function with the signature

        .. code-block:: python

           def f(params, start, end, edge, **static_kwargs):
               ...
               return x

        where the arguments are:

        - ``params``: Parameters (shared over the map)
        - ``start``: Start node data
        - ``end``: End node data
        - ``edge``: Edge data
        - ``**static_kwargs``: Any values required at compile-time
          by JAX can be passed as keyword arguments.

        The keyword argument ``key`` can be included to pass
        a random key to the mapped function.
    """
    keyword_args = utils.functions.get_keyword_args(f)
    has_key, keyword_args = utils.functions.has_key_keyword(keyword_args)

    @partial(jax.jit, static_argnames=keyword_args)
    def _edge_map(
        params: Any,
        starts: chex.ArrayTree,
        ends: chex.ArrayTree,
        edges: chex.ArrayTree,
        *,
        edge_idxs: chex.Array,
        key: Optional[chex.PRNGKey] = None,
        **static_kwargs,
    ) -> chex.ArrayTree:
        utils.functions.check_key(has_key, key)
        chex.assert_tree_has_only_ndarrays(starts)
        chex.assert_tree_has_only_ndarrays(ends)
        chex.assert_tree_has_only_ndarrays(edge_idxs)
        _check_edges(edges, edge_idxs)

        n = edge_idxs.shape[1]
        starts = jax.tree_util.tree_map(
            lambda x: x.at[edge_idxs[0]].get(),
            starts,
        )
        ends = jax.tree_util.tree_map(
            lambda x: x.at[edge_idxs[1]].get(),
            ends,
        )

        if has_key:
            keys = jax.random.split(key, n)
            results = jax.vmap(
                lambda k, pr, st, en, ed: f(pr, st, en, ed, key=k, **static_kwargs),
                in_axes=(0, None, 0, 0, 0),
            )(keys, params, starts, ends, edges)
        else:
            results = jax.vmap(partial(f, **static_kwargs), in_axes=(None, 0, 0, 0))(
                params, starts, ends, edges
            )

        return results

    return _edge_map


def graph_reduce(f: Callable, *, reduction: Reduction, n: int = -1) -> Callable:
    """
    Map function over graph edges and reduce results to nodes

    Maps the update function over the graph edges and corresponding
    node data. Then aggregates the results back to the start nodes
    of the edge-set. Edges are treated as directional, with the
    first node observing the state of the second.

    .. warning::

       Edge indices and any associated data should be
       sorted using :py:meth:`esquilax.utils.sort_edges`

    Examples
    --------

    .. testsetup:: graph_reduce

       import esquilax
       import jax
       import jax.numpy as jnp
       from functools import partial

    .. testcode:: graph_reduce

       def f(params, start, end, edge):
           return params + start + end + edge

       edge_idxs = jnp.array([[0, 0, 1], [1, 2, 0]])
       edges = jnp.array([0, 1, 2])
       starts = jnp.array([0, 1, 2])
       ends = jnp.array([0, 1, 2])

       # Call transform with edge indexes
       result = esquilax.transforms.graph_reduce(
           f,
           reduction=esquilax.reductions.add(dtype=int)
       )(
           2, starts, ends, edges, edge_idxs=edge_idxs
       )
       # result = [8, 5, 0]

    .. testcode:: graph_reduce
       :hide:

       assert result.tolist() == [8, 5, 0]

    Using :py:meth`functools.partial` the transform
    can be used as a decorator. Arguments can also be
    PyTrees, or ``None`` if unused

    .. testcode:: graph_reduce

       @partial(
           esquilax.transforms.graph_reduce,
           reduction=esquilax.reductions.add(dtype=int),
           n=3,
        )
       def f(_params, _start, end, edge):
           return end[0] + end[1] + edge

       edge_idxs = jnp.array([[0, 0, 1], [1, 2, 0]])
       edges = jnp.array([0, 1, 2])
       ends = (jnp.array([0, 1, 2]), jnp.array([0, 1, 2]))

       # Call transform with edge indexes
       f(None, None, ends, edges, edge_idxs=edge_idxs)
       # [7, 2, 0]

    .. testcode:: graph_reduce
       :hide:

       z = f(None, None, ends, edges, edge_idxs=edge_idxs)
       assert z.tolist() == [7, 2, 0]

    Random keys can be passed to the wrapped function
    by including the ``key`` keyword argument

    .. testcode:: graph_reduce

       @partial(
           esquilax.transforms.graph_reduce,
           reduction=esquilax.reductions.add(dtype=int),
           n=3,
        )
       def f(_params, _start, _end, _edge, *, key):
           return jax.random.choice(key, 100, ())

       k = jax.random.PRNGKey(101)
       result = f(
           None, None, None, None, edge_idxs=edge_idxs, key=k
       )

    Parameters
    ----------
    f
        Function with the signature

        .. code-block:: python

           def f(params, start, end, edge,  **static_kwargs):
               ...
               return x

        where the arguments are:

        - ``params``: Parameters (shared over the map)
        - ``start``: Start node data
        - ``end``: End node data
        - ``edge``: Edge data
        - ``**static_kwargs``: Any values required at compile-time
          by JAX can be passed as keyword arguments.

        The ``key`` keyword argument can be included
        to pass a random key to the mapped function.
    reduction
        Binary monoidal reduction instance
    n
        Number of nodes, should be provided if start-node data is ``None``
    """

    _edge_map = edge_map(f)
    keyword_args = utils.functions.get_keyword_args(f)
    has_key, keyword_args = utils.functions.has_key_keyword(keyword_args)

    @partial(jax.jit, static_argnames=keyword_args)
    def _graph_reduce(
        params: Any,
        starts: chex.ArrayTree,
        ends: chex.ArrayTree,
        edges: chex.ArrayTree,
        *,
        edge_idxs: chex.Array,
        key: Optional[chex.PRNGKey] = None,
        **static_kwargs,
    ) -> chex.ArrayTree:
        if starts is None:
            assert n > 0, "If starts is not provided, n should be provided"

        n_results = utils.functions.get_size(starts) if n < 0 else n

        edge_results = _edge_map(
            params,
            starts,
            ends,
            edges,
            **static_kwargs,
            edge_idxs=edge_idxs,
            key=key,
        )
        bin_counts, bins = utils.graph.index_bins(edge_idxs[0], n_results)
        bins = bins.reshape(-1)

        def reduce(r, x, d):
            x = jnp.frompyfunc(r, 2, 1).reduceat(x, bins, axis=0)[::2]
            x = jnp.where(bin_counts > 0, x, d)
            return x

        results = jax.tree_util.tree_map(
            reduce, reduction.fn, edge_results, reduction.id
        )

        return results

    return _graph_reduce


def random_edge(
    f: Callable,
    *,
    default: Default,
    n: int = -1,
) -> Callable:
    """
    Apply function to randomly selected edge from each node

    For each start node select a random neighbour on
    the graph (i.e. a random edge starting at that node)
    and apply the update function. Returns a
    default value if the node has no outgoing edges.

    .. warning::

       Edge indices and any associated data should be
       sorted using :py:meth:`esquilax.utils.sort_edges`

    Examples
    --------

    .. testsetup:: random_neighbour

       import esquilax
       import jax
       import jax.numpy as jnp
       from functools import partial

    .. testcode:: random_neighbour

       @partial(
           esquilax.transforms.random_edge,
           default=0,
           n=3,
       )
       def f(_params, _start, end, edge):
           return end[0] + end[1] + edge

       k = jax.random.PRNGKey(101)
       edge_idxs = jnp.array([[0, 0, 1], [1, 2, 0]])
       edges = jnp.array([0, 1, 2])
       ends = (jnp.array([0, 1, 2]), jnp.array([0, 1, 2]))

       # Call transform with edge indexes
       f(None, None, ends, edges, edge_idxs=edge_idxs, key=k)
       # [5, 2, 0]

    .. testcode:: random_neighbour
       :hide:

       z = f(None, None, ends, edges, edge_idxs=edge_idxs, key=k)
       assert z.tolist() == [5, 2, 0]

    Parameters
    ----------
    f
        Function with the signature

        .. code-block:: python

           def f(params, start, end, edge, **static_kwargs):
               ...
               return x

        where the arguments are:

        - ``params``: Parameters (shared over the map)
        - ``start``: Start node data
        - ``end``: End node data
        - ``edge``: Edge data
        - ``**static_kwargs``: Any values required at compile-time
          by JAX can be passed as keyword arguments.

        A random key can be passed to the wrapped function by
        including the ``key`` keyword argument.
    default
        Value returned when a node has no neighbours
    n
        Number of start nodes, should be provided if
        start-node data is None
    """

    keyword_args = utils.functions.get_keyword_args(f)
    has_key, keyword_args = utils.functions.has_key_keyword(keyword_args)

    @partial(jax.jit, static_argnames=keyword_args)
    def _random_edge(
        params: Any,
        starts: chex.ArrayTree,
        ends: chex.ArrayTree,
        edges: chex.ArrayTree,
        *,
        edge_idxs: chex.Array,
        key: chex.PRNGKey,
        **static_kwargs,
    ) -> chex.ArrayTree:
        chex.assert_tree_has_only_ndarrays(starts)
        chex.assert_tree_has_only_ndarrays(ends)
        chex.assert_tree_has_only_ndarrays(edges)
        _check_edges(edges, edge_idxs)
        if starts is None:
            assert n > 0, "If starts is not provided, n should be provided"
        n_results = utils.functions.get_size(starts) if n < 0 else n
        bin_counts, bins = utils.graph.index_bins(edge_idxs[0], n_results)

        def sample(
            _k: Optional[chex.PRNGKey], i: chex.Numeric, a: chex.Array, b: chex.Array
        ):
            k1, k2 = jax.random.split(_k)
            j = jax.random.randint(k1, (), a, b)
            end_idx = edge_idxs[1][j]
            start = jax.tree_util.tree_map(lambda x: x[i], starts)
            end = jax.tree_util.tree_map(lambda x: x[end_idx], ends)
            edge = jax.tree_util.tree_map(lambda x: x[j], edges)
            if has_key:
                result = f(params, start, end, edge, key=k2, **static_kwargs)
            else:
                result = f(params, start, end, edge, **static_kwargs)
            return result

        def select(_k, i, count: int, b: chex.Array) -> Any:
            return jax.lax.cond(
                count > 0, sample, lambda *_: default, _k, i, b[0], b[1]
            )

        keys = jax.random.split(key, n_results)
        results = jax.vmap(select, in_axes=(0, 0, 0, 0))(
            keys,
            jnp.arange(n_results),
            bin_counts,
            bins,
        )
        return results

    return _random_edge


def highest_weight(f: Callable, *, default: Default, n: int = -1) -> Callable:
    """
    Map function over graph edges with the highest weights

    For each node selects the outgoing edge with the highest
    weight, and then applies the update function on these edges
    and connected nodes. If a node has no neighbours the default value
    is returned instead.

    .. warning::

       Edge indices and any associated data should be
       sorted using :py:meth:`esquilax.utils.sort_edges`

    Examples
    --------

    .. testsetup:: graph_reduce

       import esquilax
       import jax
       import jax.numpy as jnp
       from functools import partial

    .. testcode:: graph_reduce

       def f(params, start, end, edge):
           return params + start + end + edge

       edge_idxs = jnp.array([[0, 0, 2, 2], [1, 2, 0, 1]])
       weights = jnp.array([0.1, 0.5, 0.3, 0.2])
       edges = jnp.array([0, 1, 2, 3])
       starts = jnp.array([0, 1, 2])
       ends = jnp.array([0, 1, 2])

       # Call transform with edge indexes
       result = esquilax.transforms.highest_weight(
           f, default=-1
       )(
           2,
           starts,
           ends,
           edges,
           edge_idxs=edge_idxs,
           weights=weights,
       )
       # result = [5, -1, 6]

    .. testcode:: graph_reduce
       :hide:

       assert result.tolist() == [5, -1, 6], f"{result}"

    Using :py:meth:`functools.partial` the transform can
    be used as a decorator. Arguments can also be PyTrees,
    or ``None`` if unused

    .. testcode:: graph_reduce

       @partial(
           esquilax.transforms.highest_weight,
           default=-1,
           n=3,
       )
       def f(_params, _start, end, edge):
           return end[0] + end[1] + edge

       edge_idxs = jnp.array([[0, 0, 2, 2], [1, 2, 0, 1]])
       weights = jnp.array([0.1, 0.5, 0.3, 0.2])
       edges = jnp.array([0, 1, 2, 3])
       ends = (jnp.array([0, 1, 2]), jnp.array([0, 1, 2]))

       # Call transform with edge indexes
       result = f(
           None,
           None,
           ends,
           edges,
           edge_idxs=edge_idxs,
           weights=weights,
       )
       # [5, -1, 2]

    .. testcode:: graph_reduce
       :hide:

       assert result.tolist() == [5, -1, 2], f"{z}"

    Random keys can be passed to the mapped argument by
    including the ``key`` keyword argument

    .. testcode:: graph_reduce

       def f(params, start, end, edge, *, key):
           return jax.random.choice(key, 100, ())

       k = jax.random.PRNGKey(101)
       result = esquilax.transforms.highest_weight(
           f, default=-1, n=3,
       )(
           None,
           None,
           None,
           None,
           edge_idxs=edge_idxs,
           weights=weights,
           key=k,
       )

    Parameters
    ----------
    f
        Function with the signature

        .. code-block:: python

           def f(k, params, start, end, edge,  **static_kwargs):
               ...
               return x

        where the arguments are:

        - ``params``: Parameters (shared over the map)
        - ``start``: Start node data
        - ``end``: End node data
        - ``edge``: Edge data
        - ``**static_kwargs``: Any values required at compile-time
          by JAX can be passed as keyword arguments.

        Random keys can be passed to wrapped argument by including
        the ``key`` keyword argument.
    default
        Default/identity result value
    n
        Number of nodes, should be provided if start-node data is ``None``
    """

    keyword_args = utils.functions.get_keyword_args(f)
    has_key, keyword_args = utils.functions.has_key_keyword(keyword_args)

    @partial(jax.jit, static_argnames=keyword_args)
    def _highest_weight(
        params: Any,
        starts: chex.ArrayTree,
        ends: chex.ArrayTree,
        edges: chex.ArrayTree,
        *,
        edge_idxs: chex.Array,
        weights: chex.Array,
        key: Optional[chex.PRNGKey] = None,
        **static_kwargs,
    ) -> Any:
        if starts is None:
            assert n > 0, "If starts is not provided, n should be provided"

        utils.functions.check_key(has_key, key)
        chex.assert_tree_has_only_ndarrays(starts)
        chex.assert_tree_has_only_ndarrays(ends)
        chex.assert_tree_has_only_ndarrays(edges)
        _check_edges(edges, edge_idxs)

        n_results = utils.functions.get_size(starts) if n < 0 else n
        start_nodes = edge_idxs[0]
        bin_counts, bins = utils.graph.index_bins(start_nodes, n_results)

        s0 = jnp.argsort(weights, descending=True)
        s1 = jnp.argsort(start_nodes[s0])
        s2 = s0[s1]

        max_idxs = s2[bins[:, 0]]
        max_idxs = jnp.where(bin_counts > 0, max_idxs, -1)

        def apply(k: Optional[chex.PRNGKey], i: chex.Numeric):
            e = edge_idxs[:, i]
            a = jax.tree.map(lambda x: x.at[e[0]].get(), starts)
            b = jax.tree.map(lambda x: x.at[e[1]].get(), ends)
            c = jax.tree.map(lambda x: x.at[i].get(), edges)
            if has_key:
                result = f(params, a, b, c, key=k, **static_kwargs)
            else:
                result = f(params, a, b, c, **static_kwargs)
            return result

        def check(k: Optional[chex.PRNGKey], i: chex.Numeric):
            return jax.lax.cond(i < 0, lambda *_: default, apply, k, i)

        if has_key:
            keys = jax.random.split(key, num=n_results)
            results = jax.vmap(check)(keys, max_idxs)
        else:
            results = jax.vmap(check, in_axes=(None, 0))(None, max_idxs)

        return results

    return _highest_weight



"""
Agent PyTree mapping utilities
"""
from typing import Tuple, Union

import chex
import jax
import jax.numpy as jnp

from esquilax.typing import TypedPyTree
from esquilax.utils import tree

from .agent import Agent
from .agent_state import AgentState, BatchAgentState
from .types import Trajectory


def sample_actions(
    agents: TypedPyTree[Agent],
    key: chex.PRNGKey,
    agent_states: TypedPyTree[Union[AgentState, BatchAgentState]],
    observations: chex.ArrayTree,
    greedy: bool = False,
) -> Tuple[chex.ArrayTree, chex.ArrayTree]:
    """
    Map over a tree of agents, sampling actions from observations

    Can be used to sample actions across multiple
    RL policies.

    Parameters
    ----------
    key
        JAX random key.
    agents
        Pytree of RL Agents.
    agent_states
        Pytree of corresponding agent states
    observations
        PyTree of observations, with tree structure corresponding
        to the agents.
    greedy
        Flag to indicate greedy action sampling.

    Returns
    -------
    tuple[chex.ArrayTree, chex.ArrayTree]
        Tuple containing actions, and any additional values
        associated with the actions. Both have the same
        tree structure as the argument agents.
    """
    keys = tree.key_tree_split(key, agents, typ=Agent)
    results = jax.tree.map(
        lambda agent, k, state, obs: agent.sample_actions(k, state, obs, greedy=greedy),
        agents,
        keys,
        agent_states,
        observations,
        is_leaf=lambda x: isinstance(x, Agent),
    )
    actions, action_values = tree.transpose_tree_of_tuples(agents, results, 2, Agent)
    return actions, action_values


def update_agents(
    agents: TypedPyTree[Agent],
    key: chex.PRNGKey,
    agent_states: TypedPyTree[Union[AgentState, BatchAgentState]],
    trajectories: TypedPyTree[Trajectory],
) -> Tuple[TypedPyTree[Agent], chex.ArrayTree]:
    """
    Update agent states from gathered trajectories

    Parameters
    ----------
    key
        JAX random key.
    agents
        PyTree of RL-agents.
    agent_states
        Pytree of corresponding agent states
    trajectories
        PyTree of environment trajectories.

    Returns
    -------
    tuple[esquilax.typing.TypedPyTree[esquilax.ml.rl.AgentState], chex.ArrayTree]
        Tuple containing PyTrees of updated
        agents, and any data returned from training
        (e.g. training loss). Both trees have the same
        structure as the argument agents.
    """
    keys = tree.key_tree_split(key, agents, typ=Agent)

    updates = jax.tree.map(
        lambda agent, k, state, traj: agent.update(k, state, traj),
        agents,
        keys,
        agent_states,
        trajectories,
        is_leaf=lambda x: isinstance(x, Agent),
    )
    agents, train_data = tree.transpose_tree_of_tuples(agents, updates, 2, Agent)
    return agents, train_data


def reshape_trajectories(trajectories: Trajectory) -> Trajectory:
    """
    Reshape a batch of trajectories into individual agent histories

    Reshape batch of trajectories gathered across
    multiple environments, i.e. with shape
    ``[n-envs, n_steps, n-agents, ...]`` into trajectory
    histories per agent and environment, i.e. with shape
    ``[n-envs * n-agents, n_steps, ...]``.

    Parameters
    ----------
    trajectories
        Batch of trajectories with array shapes
        ``[n-envs, n_steps, n-agents, ...]``.

    Returns
    -------
    esquilax.ml.rl.Trajectory
        Reshaped trajectories with shape
        ``[n-envs * n-agents, n_steps, ...]``.
    """

    def reshape(x):
        x = jnp.swapaxes(x, 1, 2)
        x = jnp.reshape(x, (x.shape[0] * x.shape[1],) + x.shape[2:])
        return x

    trajectories = jax.tree.map(reshape, trajectories)

    return trajectories



"""
Simulation update utilities for RL and neuro-evolution
"""
from functools import partial
from typing import Callable

import chex
import jax


@partial(jax.jit, static_argnames=("f", "broadcast"))
def get_actions(
    f: Callable, broadcast: bool, params: chex.ArrayTree, observations: chex.ArrayTree
) -> chex.ArrayTree:
    """
    Apply a function and parameters to observations

    Generate actions from a function and corresponding
    parameters applied to observation. Typically, this
    may be a neural-network forward pass, with network
    parameters, applied to observation of some
    simulation state.

    This is intended to sample actions across multiple
    agents, with either shared parameters, or parameters
    per agent (each with corresponding individual
    observations),

    Parameters
    ----------
    f
        Function parameterised by ``params`` (e.g.
        a neural-network apply function). Should
        have a signature :code:`(params, observation) -> x`.
    broadcast
        If ``True`` the parameters will be shared across
        all the observations (i.e. a shared policy),
        otherwise they will be mapped over.
    params
        Function parameters. Since shared
        by agents these should be a single sample
        of parameters.
    observations
        Array/tree of individual observation for agents
        to be mapped over.

    Returns
    -------
    chex.ArrayTree
        Output of ``f`` for each agent
    """
    in_axes = (None, 0) if broadcast else (0, 0)
    return jax.vmap(f, in_axes=in_axes)(params, observations)


@partial(jax.jit, static_argnames=("f", "broadcast"))
def sample_actions(
    f: Callable,
    broadcast: bool,
    key: chex.PRNGKey,
    params: chex.ArrayTree,
    observations: chex.ArrayTree,
) -> chex.ArrayTree:
    """
    Apply a function and parameters to observations with a random key

    Generate actions from a function and corresponding
    parameters applied to observation. Also passes
    a random key for use when some randomness is
    required, e.g. for policy exploration during RL training.

    Parameters
    ----------
    f
        Function parameterised by ``params`` (e.g.
        a neural-network apply function). Should
        have a signature :code:`(key, params, observation) -> x`.
    broadcast
        If ``True`` the parameters will be shared across
        all the observations (i.e. a shared policy),
        otherwise they will be mapped over.
    key
        JAX random key.
    params
        Function parameters. Since shared
        by agents these should be a single sample
        of parameters.
    observations
        Array/tree of individual observation for agents
        to be mapped over.

    Returns
    -------
    chex.ArrayTree
        Output of ``f`` for each agent.
    """
    in_axes = (0, None, 0) if broadcast else (0, 0, 0)
    keys = jax.random.split(key, observations.shape[0])
    return jax.vmap(f, in_axes=in_axes)(keys, params, observations)
